# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

__all__ = ['predictors', 'uncertain_data', 'state_estimators', 'run_prog_playback']
from .predictors import *
from .state_estimators import *
from .uncertain_data import *

import numpy as np

def run_prog_playback(obs, pred, future_loading, output_measurements, **kwargs):
    config = {# Defaults
        'predict_rate': 0, # Default- predict every step
        'num_samples': 10,
        'predict_config': {}
    }
    config.update(kwargs)

    next_predict = output_measurements[0][0] + config['predict_rate']
    times = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    inputs = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    states = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    outputs = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    event_states = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    eols = np.empty((len(output_measurements), config['num_samples']), dtype=object)
    index = 0
    for (t, measurement) in output_measurements:
        obs.estimate(t, future_loading(t), measurement)
        if t >= next_predict:
            (t, u, x, z, es, eol) = pred.predict(obs.x.sample(config['num_samples']), future_loading, **config['predict_config'])
            times[index, :] = t
            inputs[index, :]  = u
            states[index, :]  = x
            outputs[index, :]  = z
            event_states[index, :]  =  es
            eols[index, :] = eol
            index += 1
            next_predict += config['predict_rate']
    return (times, inputs, states, outputs, event_states, eols) 



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

from abc import ABC, abstractmethod, abstractproperty
from numpy.random import choice, multivariate_normal
from numpy import array, append, delete, cov

class UncertainData(ABC):
    """
    Abstract base class for data with uncertainty. Any new uncertainty type must implement this class
    """
    @abstractmethod
    def sample(self, nSamples = 1):
        """Generate samples from data

        Args:
            nSamples (int, optional): Number of samples to generate. Defaults to 1.

        Returns:
            samples (array): Array of nSamples samples

        Example:
            samples = data.samples(100)
        """
        pass

    @property
    @abstractproperty
    def mean(self):
        """Mean estimate

        Example:
            mean_value = data.mean
        """
        pass

    @property
    @abstractproperty
    def cov(self):
        """Get the covariance matrix

        Returns:
            [[float]]: covariance matrix
        """

    # TODO(CT): Consider median

class ScalarData(UncertainData):
    """
    Data without uncertainty- single value
    """
    def __init__(self, state): 
        """Initialize Scalar Data

        Args:
            state (dict): Single state
        """
        self.__state = state

    @property
    def mean(self):
        return self.__state

    @property
    def cov(self):
        return [[0 for i in self.__state] for j in self.__state]

    def sample(self, num_samples = 1):
        return array([self.__state] * num_samples)

    def __str__(self):
        return 'ScalarData({})'.format(self.__state)


class UnweightedSamples(UncertainData):
    """
    Uncertain Data represented by a set of samples
    """
    def __init__(self, samples : array = array([])):
        """Initialize Unweighted Samples

        Args:
            samples (array, optional): array of samples. Defaults to empty array.
        """
        self.__samples = samples

    def sample(self, num_samples = 1):
        # Completely random resample
        return choice(self.__samples, num_samples)

    @property
    def mean(self):
        mean = {}
        for key in self.__samples[0].keys():
            mean[key] = array([x[key] for x in self.__samples]).mean()
        return mean

    @property
    def cov(self):
        if len(self.__samples) == 0:
            return [[]]
        unlabeled_samples = array([[x[key] for x in self.__samples] for key in self.__samples[0].keys()])
        return cov(unlabeled_samples)

    def __str__(self):
        return 'UnweightedSamples({})'.format(self.__samples)

    # Sample-specific methods
    def append(self, value):
        """Append an additional sample to the unweighted samples

        Args:
            value (dict): Value to be appended
        """
        self.__samples = append(self.__samples, value)

    @property
    def size(self):
        """Get the number of samples

        Returns:
            int: Number of samples
        """
        return self.__samples.size

    def __getitem__(self, index):
        """Get a specific item by index

        Args:
            index (int): Sample index requested

        Returns:
            dict: item requested

        Example:
            sample = samples[index]
        """
        return self.__samples[index]

    def __setitem__(self, index, value):
        """Set a specific item by index

        Args:
            index (int): index to be set
            value (dict): new value

        Example:
            samples[index] = new_value
        """
        self.__samples[index] = value

    def __delitem__(self, index):
        """Delete a specific utem

        Args:
            index (int): Index of item to be deleted

        Example:
            del samples[index]
        """
        self.__samples = delete(self.__samples, index)

    def raw_samples(self):
        """Get raw samples

        Returns:
            np.array(dict): all the samples
        """
        return self.__samples

class MultivariateNormalDist(UncertainData):
    """
    Data represented by a multivariate normal distribution with mean and covariance matrix
    """
    def __init__(self, labels, mean: array, covar : array):
        """Initialize distribution

        Args:
            labels ([str]): Labels for states, in order of mean values
            mean ([float]): Mean values for state in the same order as labels
            covar ([[float]]): Covariance matrix for state
        """
        self.__labels = labels
        self.__mean = mean
        self.__covar = covar

    def sample(self, num_samples = 1):
        if len(self.__mean) != len(self.__labels):
            raise Exception("labels must be provided for each value")
    
        samples = multivariate_normal(self.__mean, self.__covar, num_samples)
        samples = array([{key: value for (key, value) in zip(self.__labels, x)} for x in samples])
        return samples

    @property
    def mean(self):
        return {key: value for (key, value) in zip(self.__labels, self.__mean)}

    def __str__(self):
        return 'MultivariateNormalDist(mean: {}, covar: {})'.format(self.__mean, self.__covar)     

    @property
    def cov(self):
        return self.__covar
    


# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

__all__ = ['state_estimator', 'unscented_kalman_filter', 'particle_filter']


# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from . import state_estimator
from numpy import array, empty, random
import filterpy.monte_carlo
from numbers import Number
from scipy.stats import norm
from ..uncertain_data import UnweightedSamples
from ..exceptions import ProgAlgTypeError
from copy import deepcopy

class ParticleFilter(state_estimator.StateEstimator):
    """
    Estimates state using a particle filter algorithm

    Constructor parameters:
     * model (ProgModel): Model to be used in state estimation \n
        See: Prognostics Model Package \n
        A prognostics model to be used in state estimation
     * x0 (dict): Initial State \n
        Initial (starting) state, with keys defined by model.states \n
        e.g., x = {'abc': 332.1, 'def': 221.003} given states = ['abc', 'def']
     * measurement_eqn (optional, function): Measurement equation (x)->z. Usually used in situations where what's measured don't exactly match the output (e.g., different unit, not ever output measured, etc.). see `examples.measurement_eqn_example`
     * options (optional, kwargs): configuration options\n
        Any additional configuration values. See default parameters. Additionally, the following configuration parameters are supported: \n
         * num_particles : Number of particles used in PF e.g., 20
         * resample_fcn : Resampling function ([weights]) -> [indexes] e.g., filterpy.monte_carlo.residual_resample
         * x0_uncertainty : Initial uncertainty in state e.g., 0.5
         * R (Number) : Measurement Noise. e.g., 0.1
    """
    t = 0 # last timestep

    default_parameters = {
            'n': 0.1, # Sensor Noise
            'num_particles': 20, 
            'resample_fcn': filterpy.monte_carlo.residual_resample, # Resampling function ([weights]) -> [indexes]
            'x0_uncertainty': 0.5   # Initial State Uncertainty
                                    # Can be:
                                    #   1. scalar (standard deviation applied to all),
                                    #   2. dict (stardard deviation for each)
                                    #   Todo(CT): covar, function
        }

    def __init__(self, model, x0, measurement_eqn = None, **kwargs):
        super().__init__(model, x0, measurement_eqn = measurement_eqn, **kwargs)

        if measurement_eqn is None:
            self.__measure = model.output
        else:
            self.__measure = measurement_eqn

        # State-estimator specific logic
        if isinstance(self.parameters['n'], Number):
            self.parameters['n'] = {key : self.parameters['n'] for key in self.__measure(x0).keys()}
        # todo(CT): Check fields on n

        if isinstance(self.parameters['x0_uncertainty'], Number):
            # Build array inplace
            x = array(list(x0.values()))
            sd = array([self.parameters['x0_uncertainty']] * len(x0))
            samples = array([random.normal(x, sd) for i in range(self.parameters['num_particles'])])
            self.particles = array([{key: value for (key, value) in zip(model.states, x)} for x in samples])
        else:
            raise ProgAlgTypeError
    
    def __str__(self):
        return "{} State Estimator".format(self.__class__)
        
    def estimate(self, t, u, z):
        # todo(CT): assert t > self.t?
        dt = t - self.t
        weights = empty(len(self.particles))
        
        # Optimization
        particles = self.particles
        next_state = self.model.next_state
        output = self.__measure
        noise_params = self.parameters['n']

        # Propogate and calculate weights
        for i in range(len(particles)):
            self.particles[i] = next_state(particles[i], u, dt) 
            zPredicted = output(self.particles[i])
            weights[i] = sum([norm(zPredicted[key], noise_params[key]).pdf(z[key]) for key in zPredicted.keys()])
        
        # Normalize
        total_weight = sum(weights)
        weights = array([weight/total_weight for weight in weights])

        # Resample
        indexes = self.parameters['resample_fcn'](weights)
        self.particles = array([self.particles[i] for i in indexes])

    @property
    def x(self):
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        return UnweightedSamples(self.particles)



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from abc import ABC, abstractmethod, abstractproperty
from ..uncertain_data import UncertainData
from ..exceptions import ProgAlgTypeError
from copy import deepcopy

class StateEstimator(ABC):
    """
    Interface class for state estimators

    Abstract base class for creating state estimators that perform state estimation. Subclasses must implement this interface. Equivilant to "Observers" in NASA's Matlab Prognostics Algorithm Library

    Constructor parameters:
     * model (ProgModel): Model to be used in state estimation \n
        See: Prognostics Model Package \n
        A prognostics model to be used in state estimation
     * x0 (dict): Initial State \n
        Initial (starting) state, with keys defined by model.states \n
        e.g., x = {'abc': 332.1, 'def': 221.003} given states = ['abc', 'def']
     * measurement_eqn (optional, function): Measurement equation (x)->z. Usually used in situations where what's measured don't exactly match the output (e.g., different unit, not ever output measured, etc.). see `examples.measurement_eqn_example`
     * options (optional, kwargs): configuration options\n
        Dictionary of any additional configuration values. See state-estimator specific documentation
    """

    default_parameters = {}

    def __init__(self, model, x0, measurement_eqn = None, **kwargs):
        # Check model
        if not hasattr(model, 'output'):
            raise ProgAlgTypeError("model must have `output` method")
        if not hasattr(model, 'next_state'):
            raise ProgAlgTypeError("model must have `next_state` method")
        if not hasattr(model, 'outputs'):
            raise ProgAlgTypeError("model must have `outputs` property")
        if not hasattr(model, 'states'):
            raise ProgAlgTypeError("model must have `states` property")
        self.model = model

        # Check x0
        for key in model.states:
            if key not in x0:
                raise ProgAlgTypeError("x0 missing state `{}`".format(key))
        
        # Check measurement equation
        if measurement_eqn and not callable(measurement_eqn):
            raise ProgAlgTypeError("measurement_eqn must be callable")
        
        # Process kwargs (configuration)
        self.parameters = deepcopy(self.default_parameters)
        self.parameters.update(kwargs)

    @abstractmethod
    def estimate(self, t, u, z):
        """
        Perform one state estimation step (i.e., update the state estimate)

        Parameters
        ----------
        t : double
            Current timestamp in seconds (≥ 0.0)
            e.g., t = 3.4
        u : dict
            Measured inputs, with keys defined by model.inputs.
            e.g., u = {'i':3.2} given inputs = ['i']
        z : dict
            Measured outputs, with keys defined by model.outputs.
            e.g., z = {'t':12.4, 'v':3.3} given inputs = ['t', 'v']
        """
        pass

    @property
    @abstractproperty
    def x(self) -> UncertainData:
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        pass



# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration. All Rights Reserved.

class ProgAlgException(Exception):
    """
    Base Prognostics Model Exception
    """
    pass

class ProgAlgInputException(ProgAlgException):
    """
    Prognostics Input Exception - indicates the method input parameters were incorrect
    """
    pass

class ProgAlgTypeError(ProgAlgException, TypeError):
    """
    Prognostics Type Error - indicates the model could not be constructed
    """
    pass


# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from . import state_estimator
from filterpy import kalman
from numpy import diag, array
from ..uncertain_data import MultivariateNormalDist
from ..exceptions import ProgAlgTypeError
from copy import deepcopy

class UnscentedKalmanFilter(state_estimator.StateEstimator):
    """
    An Unscented Kalman Filter (UKF) for state estimation

    This class defines logic for performing an unscented kalman filter with a Prognostics Model (see Prognostics Model Package). This filter uses measurement data with noise to generate a state estimate and covariance matrix. 
    
    Constructor parameters:
     * model (ProgModel): Model to be used in state estimation \n
        See: Prognostics Model Package \n
        A prognostics model to be used in state estimation
     * x0 (dict): Initial State \n
        Initial (starting) state, with keys defined by model.states \n
        e.g., x = {'abc': 332.1, 'def': 221.003} given states = ['abc', 'def']
     * measurement_eqn (optional, function): Measurement equation (x)->z. Usually used in situations where what's measured don't exactly match the output (e.g., different unit, not ever output measured, etc.). see `examples.measurement_eqn_example`
     * options (optional, kwargs): configuration options\n
        Dictionary of any additional configuration values. See default parameters. Additionally, the following configuration parameters are supported: \n
         * alpha, beta, kappa: UKF Scaling parameters
         * t0 : Starting time
         * dt : time step
         * Q : Process Noise Matrix 
         * R : Measurement Noise Matrix 
    """
    t = 0 # Last timestep
    default_parameters = {
        'alpha': 1,     # UKF scaling param
        'beta': 0,      # UKF scaling param
        'kappa': -1,    # UKF scaling param
        't0': 0,        # Starting time
        'dt': 1         # Time step
    } 

    def __init__(self, model, x0, measurement_eqn = None, **kwargs):
        super().__init__(model, x0, measurement_eqn, **kwargs)

        self._input = None
        self.t = self.parameters['t0']

        if measurement_eqn is None: 
            def measure(x):
                x = {key: value for (key, value) in zip(model.states, x)}
                z = model.output(x)
                return array(list(z.values()))
        else:
            def measure(x):
                x = {key: value for (key, value) in zip(model.states, x)}
                z = measurement_eqn(x)
                return array(list(z.values()))

        if 'Q' not in self.parameters:
            self.parameters['Q'] = diag([1.0e-1 for i in model.states])
        if 'R' not in self.parameters:
            self.parameters['R'] = diag([1.0e-1 for i in range(len(measure(x0.values())))])

        def state_transition(x, dt):
            x = {key: value for (key, value) in zip(model.states, x)}
            x = model.next_state(x, self._input, dt)
            return array(list(x.values()))

        num_states = len(model.states)
        num_measurements = len(model.outputs)
        points = kalman.MerweScaledSigmaPoints(num_states, alpha=self.parameters['alpha'], beta=self.parameters['beta'], kappa=self.parameters['kappa'])
        self.filter = kalman.UnscentedKalmanFilter(num_states, num_measurements, self.parameters['dt'], measure, state_transition, points)
        self.filter.x = array(list(x0.values()))
        self.filter.Q = self.parameters['Q']
        self.filter.R = self.parameters['R']

    def estimate(self, t, u, z):
        """
        Perform one state estimation step (i.e., update the state estimate)

        Parameters
        ----------
        t : double
            Current timestamp in seconds (≥ 0.0)
            e.g., t = 3.4
        u : dict
            Measured inputs, with keys defined by model.inputs.
            e.g., u = {'i':3.2} given inputs = ['i']
        z : dict
            Measured outputs, with keys defined by model.outputs.
            e.g., z = {'t':12.4, 'v':3.3} given inputs = ['t', 'v']
        """
        dt = t - self.t
        self._input = u
        self.t = t
        self.filter.predict(dt=dt)
        self.filter.update(array(list(z.values())))
    
    @property
    def x(self):
        """
        Getter for property 'x', the current estimated state. 

        Example
        -------
        state = observer.x
        """
        return MultivariateNormalDist(self.model.states, self.filter.x, self.filter.Q)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name = 'prog_algs',
    version = '0.3.0',
    description = "The NASA Prognostic Algorithm Package is a framework for model-based prognostics (computation of remaining useful life) of engineering systems. It includes algorithms for state estimation and prediction, including uncertainty propagation. The algorithms use prognostic models (see prog_models) to perform estimation and prediction. The package enables rapid development of prognostics solutions for given models of components and systems. Algorithms can be swapped for comparative studies and evaluations",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url = 'https://github.com/nasa/prog_algs',
    author = 'Christopher Teubert',
    author_email = 'christopher.a.teubert@nasa.gov',
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Intended Audience :: Manufacturing', 
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: Other/Proprietary License ',   
        'Programming Language :: Python :: 3',     
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only'
    ],
    keywords = ['prognostics', 'diagnostics', 'fault detection', 'fdir', 'prognostics and health management', 'PHM', 'health management'],
    packages = ['prog_algs', 'prog_algs.metrics', 'prog_algs.state_estimators', 'prog_algs.uncertain_data', 'prog_algs.visualize'],
    python_requires='>=3.6, <3.9',
    install_requires = [
        'numpy',
        'scipy',
        'filtpy',
        'matplotlib'
    ],
    license = 'NOSA',
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/nasa/prog_algs/issues',
        'Organization': 'https://prognostics.nasa.gov/',
        'Source': 'https://github.com/nasa/prog_algs',
    }
)


# Copyright © 2021 United States Government as represented by the Administrator of the National Aeronautics and Space Administration.  All Rights Reserved.

from numpy import mean, std
from scipy import stats

def eol_metrics(eol, ground_truth = None):
    """Calculate all end of life metrics

    Args:
        eol ([double]): End of life times, output from predictor
        ground_truth (float, optional): Ground truth end of discharge time. Defaults to None.

    Returns:
        dict: collection of metrics
    """
    eol.sort()
    m = mean(eol)
    median = eol[int(len(eol)/2)]
    metrics = {
        'min': eol[0],
        'percentiles': {
            '0.01': eol[int(len(eol)/10000)] if len(eol) >= 10000 else None,
            '0.1': eol[int(len(eol)/1000)] if len(eol) >= 1000 else None,
            '1': eol[int(len(eol)/100)] if len(eol) >= 100 else None,
            '10': eol[int(len(eol)/10)] if len(eol) >= 10 else None,
            '25': eol[int(len(eol)/4)] if len(eol) >= 4 else None,
            '50': median,
            '75': eol[int(3*len(eol)/4)] if len(eol) >= 4 else None,
        },
        'median': median,
        'mean': m,
        'std': std(eol),
        'max': eol[-1],
        'median absolute deviation': sum([abs(x - median) for x in eol])/len(eol),
        'mean absolute deviation':   sum([abs(x - m)   for x in eol])/len(eol),
        'number of samples': len(eol)
    }

    if ground_truth is not None:
        # Metrics comparing to ground truth
        metrics['mean absolute error'] = sum([abs(x - ground_truth) for x in eol])/len(eol)
        metrics['mean absolute percentage error'] = metrics['mean absolute error']/ ground_truth
        metrics['relative accuracy'] = 1 - abs(ground_truth - metrics['mean'])/ground_truth
        metrics['ground truth percentile'] = stats.percentileofscore(eol, ground_truth)

    return metrics

def prob_success(eol, time):
    """Calculate probability of success - i.e., probability that event will not occur within a given time (i.e., success)

    Args:
        eol ([float]): End of life times, output from predictor
        time ([type]): time for calculation

    Returns:
        fload: Probability of success
    """
    return sum([e > time for e in eol])/len(eol)

def alpha_lambda(times, eols, ground_truth, lambda_value, alpha, beta): 
    """Compute alpha lambda metrics

    Args:
        times ([float]): Times corresponding to eols (output from predictors)
        eols ([float]): End of life times, output from predictor
        ground_truth (float): Ground Truth end of life time
        lambda_value (fload): lambda
        alpha (float): alpha
        beta (float): beta

    Returns:
        bool: if alpha-lambda met
    """
    for (t, eol) in zip(times, eols):
        if (t >= lambda_value):
            upper_bound = ground_truth + alpha*(ground_truth-t)
            lower_bound = ground_truth - alpha*(ground_truth-t)
            return percentage_in_bounds(eol, [lower_bound, upper_bound]) >= beta 


def mean_square_error(values, ground_truth):
    """Mean Square Error

    Args:
        values ([float]): End of life times, output from predictor
        ground_truth (float): Ground truth EOL

    Returns:
        float: mean square error of eol predictions
    """
    return sum([(x.mean() - ground_truth)**2 for x in values])/len(values)

def eol_profile_metrics(eol, ground_truth):
    """Calculate eol profile metrics

    Args:
        eol ([float]): End of life times, output from predictor
        ground_truth (float): Ground truth EOL

    Returns:
        dict: EOL Profile Metrics
    """
    # TODO(CT): Consider ground truth optional
    return {
        'mean square error': mean_square_error(eol, ground_truth)
    }

def percentage_in_bounds(eol, bounds):
    """Calculate percentage of EOL dist is within specified bounds

    Args:
        eol ([float]): End of life times, output from predictor
        bounds ([float, float]): Lower and upper bounds

    Returns:
        float: Percentage within bounds (where 1 = 100%)
    """
    return sum([x < bounds[1] and x > bounds[0] for x in eol])/ len(eol)